package com.bcxin.tenant.open.dubbo.reader.providers.impls;

import com.bcxin.tenant.open.document.domains.documents.*;
import com.bcxin.tenant.open.document.domains.documents.RdCompanyDocument;
import com.bcxin.tenant.open.document.domains.documents.RdCompanyDocument$;
import com.bcxin.tenant.open.document.domains.repositories.RdCompanyDocumentRepository;
import com.bcxin.tenant.open.document.domains.utils.KeywordTermUtils;
import com.bcxin.tenant.open.dubbo.reader.providers.transfers.DataTransfer;
import com.bcxin.tenant.open.infrastructure.redis.RedisConfig;
import com.bcxin.tenant.open.infrastructures.EntityCollection;
import com.bcxin.tenant.open.infrastructures.TenantContext;
import com.bcxin.tenant.open.infrastructures.TenantEmployeeContext;
import com.bcxin.tenant.open.infrastructures.enums.DeskType;
import com.bcxin.tenant.open.infrastructures.enums.ResourceType;
import com.bcxin.tenant.open.infrastructures.exceptions.ArgumentTenantException;
import com.bcxin.tenant.open.infrastructures.exceptions.UnAuthorizedTenantException;
import com.bcxin.tenant.open.infrastructures.utils.GisPointUtil;
import com.bcxin.tenant.open.infrastructures.valueTypes.LonLatValueType;
import com.bcxin.tenant.open.jdks.CompanyReaderRpcProvider;
import com.bcxin.tenant.open.jdks.requests.CompanyOverviewRequest;
import com.bcxin.tenant.open.jdks.requests.CompanySearchRequest;
import com.bcxin.tenant.open.jdks.requests.SearchNearByCompanyRequest;
import com.bcxin.tenant.open.jdks.responses.CompanyDetailResponse;
import com.bcxin.tenant.open.jdks.responses.CompanyAdvanceResponse;
import com.bcxin.tenant.open.jdks.responses.CompanySearchResponse;
import com.bcxin.tenant.open.jdks.responses.SearchNearByCompanyResponse;
import com.bcxin.tenant.open.infrastructures.utils.BusinessUtil;
import com.bcxin.tenant.open.jdks.CompanyReaderRpcProvider;
import com.bcxin.tenant.open.jdks.requests.CompanyOverviewRequest;
import com.bcxin.tenant.open.jdks.requests.CompanySearchRequest;
import com.bcxin.tenant.open.jdks.requests.EmployeeSearchRequest;
import com.bcxin.tenant.open.jdks.responses.CompanyDetailResponse;
import com.bcxin.tenant.open.jdks.responses.CompanyAdvanceResponse;
import com.bcxin.tenant.open.jdks.responses.CompanySearchResponse;
import com.redis.om.spring.annotations.ReducerFunction;
import com.redis.om.spring.metamodel.MetamodelField;
import com.redis.om.spring.search.stream.AggregationStream;
import com.redis.om.spring.search.stream.EntityStream;
import com.redis.om.spring.search.stream.SearchStream;
import com.redis.om.spring.tuple.Pair;
import com.redis.om.spring.tuple.Triple;
import com.redis.om.spring.tuple.impl.QuadImpl;
import org.apache.dubbo.config.annotation.DubboService;
import org.springframework.data.domain.Sort;
import org.springframework.data.geo.Distance;
import org.springframework.data.geo.Point;
import org.springframework.data.redis.domain.geo.Metrics;
import com.redis.om.spring.tuple.Triple;
import org.apache.dubbo.config.annotation.DubboService;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@DubboService
public class CompanyReaderRpcProviderImpl implements CompanyReaderRpcProvider {
    private final RdCompanyDocumentRepository companyDocumentRepository;
    private final DataTransfer dataTransfer;
    private final EntityStream entityStream;

    public CompanyReaderRpcProviderImpl(RdCompanyDocumentRepository companyDocumentRepository, DataTransfer dataTransfer, EntityStream entityStream) {
        this.companyDocumentRepository = companyDocumentRepository;
        this.dataTransfer = dataTransfer;
        this.entityStream = entityStream;
    }

    @Override
    public CompanyDetailResponse get(String id, boolean includeForExamSiteRoom) {
        Optional<RdCompanyDocument> documentOptional =
                companyDocumentRepository.findById(id);

        if (!documentOptional.isPresent()) {
            return null;
        }

        RdCompanyDocument document = documentOptional.get();

        Integer countOfExamSiteRooms = null;
        if (includeForExamSiteRoom) {
            SearchStream<RdExamSiteRoomDocument> examSiteRoomDocumentSearchStream =
                    this.entityStream.of(RdExamSiteRoomDocument.class);
            examSiteRoomDocumentSearchStream = examSiteRoomDocumentSearchStream
                    .filter(RdExamSiteRoomDocument$.ORGANIZATION_ID.eq(document.getOrganizationId()));
            examSiteRoomDocumentSearchStream = examSiteRoomDocumentSearchStream.filter(RdExamSiteRoomDocument$.STATUS.eq(true));

            countOfExamSiteRooms = (int) examSiteRoomDocumentSearchStream.count();
        }

        return this.dataTransfer.translate(document, countOfExamSiteRooms);
    }

    @Override
    public Collection<CompanySearchResponse> search(CompanySearchRequest request) {
        if (CollectionUtils.isEmpty(request.getDeskTypes())) {
            throw new ArgumentTenantException("参数异常; 工作台类型(deskTypes)不能为空");
        }

        if (request.getSearchModel() == null) {
            throw new ArgumentTenantException("参数异常, 搜索模式不能为空");
        }

        if (request.getSearchModel() == CompanySearchRequest.SearchModel.SecurityResource) {
            if (CollectionUtils.isEmpty(request.getInstitutionals())) {
                throw new ArgumentTenantException("参数异常; 单位类型参数不能为空");
            }
        }

        TenantEmployeeContext.TenantUserModel userModel = TenantContext.getInstance().getUserContext().get();
        /**
         * 根据监管归属进行过滤, 暂时不支持企业级的调度台过滤
         */
        SearchStream<RdCompanyDocument> companyDocumentSearchStream =
                this.entityStream.of(RdCompanyDocument.class);

        Set<String> selectedResources
                = ResourceType.getResources(request.getDeskTypes());
        if (!CollectionUtils.isEmpty(selectedResources)) {
            companyDocumentSearchStream =
                    companyDocumentSearchStream.filter(RdEmployeeDocument$.RESOURCE_TYPES.in(
                            selectedResources.toArray(ix -> new String[ix])
                    ));
        }

        if (userModel.isSuperviseDepartRole()) {
            String[] superviseDepartIds =
                    userModel.getMatchedSuperviseDepartIds(request.getSuperviseDepartIds())
                            .toArray(ii -> new String[ii]);
            if (superviseDepartIds == null || superviseDepartIds.length == 0) {
                return Collections.EMPTY_LIST;
            }

            /**
             * 根据监管归属进行过滤, 暂时不支持企业级的调度台过滤
             */
            companyDocumentSearchStream = companyDocumentSearchStream
                    .filter(RdCompanyDocument$.SCOPE_PERMISSIONS.in(superviseDepartIds));

            if (!CollectionUtils.isEmpty(request.getInstitutionals())) {
                companyDocumentSearchStream = companyDocumentSearchStream
                        .filter(RdCompanyDocument$.INSTITUTIONAL.in(request.getInstitutionals().toArray(ix -> new String[ix])));
            }
        } else {
            /**
             * 本企业(及子公司)调度
             */
            companyDocumentSearchStream = companyDocumentSearchStream
                    .filter(RdCompanyDocument$.ID.in(userModel.getNoSuperviseScopePermissions()));

            if (!CollectionUtils.isEmpty(request.getSuperviseDepartIds())) {
                companyDocumentSearchStream = companyDocumentSearchStream
                        .filter(RdCompanyDocument$.SCOPE_PERMISSIONS.in(request.getSuperviseDepartIds().toArray(ii -> new String[ii])));
            }

            /**
             * 针对企业调度台; 先临时放着; 如果机构类型刚好是非02/03/04; 那么过滤将是无效的
             */
            if (!CollectionUtils.isEmpty(request.getInstitutionals())) {
                companyDocumentSearchStream = companyDocumentSearchStream
                        .filter(RdCompanyDocument$.INSTITUTIONAL.in(request.getInstitutionals().toArray(ix -> new String[ix])));
            }
        }

        if (StringUtils.hasLength(request.getKeyword())) {
            companyDocumentSearchStream = filterByKeyword(companyDocumentSearchStream, request.getKeyword());
        }

        Collection<RdCompanyDocument> companyDocuments = null;
        /**
         * 对于没有分页的限制最多100万条数据
         */
        if (request.getSearchModel() == CompanySearchRequest.SearchModel.Keyword) {
            int pageIndex = request.getPageIndex();
            if (pageIndex < 1) {
                pageIndex = 1;
            }

            companyDocuments = companyDocumentSearchStream
                    /**
                     * 这个一定要再map之前, 否则无法执行
                     */
                    .skip((pageIndex - 1) * request.getPageSize())
                    .limit(request.getPageSize())
                    .collect(Collectors.toList());
        } else {
            if (request.getFollowed() != null) {
                if (request.getFollowed()) {
                    companyDocumentSearchStream =
                            companyDocumentSearchStream.filter(
                                    RdEmployeeDocument$.FOLLOWED_DEVICE_NOS.in(userModel.getTencentUserId())
                            );
                } else {
                    /**
                     * 关注人不在列表中
                     */
                    Set<String> noDeviceNos = new HashSet<>();
                    noDeviceNos.add(userModel.getTencentUserId());

                    companyDocumentSearchStream =
                            companyDocumentSearchStream.filter(
                                    RdEmployeeDocument$.FOLLOWED_DEVICE_NOS.containsNone(noDeviceNos)
                            );
                }
            }

            companyDocuments = companyDocumentSearchStream
                    /**
                     * 这个一定要再map之前, 否则无法执行
                     */
                    .limit(RedisConfig.MAX_LIMIT)
                    .collect(Collectors.toList());
        }


        return companyDocuments.stream().map(rdCompanyDocument -> {
            boolean follow = false;
            if (request.getSearchModel() != CompanySearchRequest.SearchModel.SecurityResource) {
                if (rdCompanyDocument.getFollowedDeviceNos() != null) {
                    if (userModel != null) {
                        follow = rdCompanyDocument.getFollowedDeviceNos().stream().anyMatch(ii ->
                                StringUtils.hasLength(ii) && ii.equalsIgnoreCase(userModel.getTencentUserId())
                        );
                    }
                }
            }

            return CompanySearchResponse.create(
                    request.getSearchModel(),
                    rdCompanyDocument.getId(),
                    rdCompanyDocument.getName(),
                    rdCompanyDocument.getLegal(),
                    rdCompanyDocument.getLegalTelephone(),
                    rdCompanyDocument.getRegisterAddress(),
                    rdCompanyDocument.getBusinessAddress(),
                    rdCompanyDocument.getLonLat().getY(),
                    rdCompanyDocument.getLonLat().getX(),
                    rdCompanyDocument.getFollowedDeviceNos(),
                    follow,
                    rdCompanyDocument.getCountOfTemporaryProject()
            );
        }).collect(Collectors.toList());
    }

    @Override
    public EntityCollection<CompanyAdvanceResponse> search(CompanyOverviewRequest request) {
        TenantEmployeeContext.TenantUserModel userModel = TenantContext.getInstance().getUserContext().get();
        if (userModel == null) {
            throw new UnAuthorizedTenantException();
        }

        SearchStream<RdCompanyDocument> searchStream =
                this.entityStream.of(RdCompanyDocument.class);
        if (userModel.isSuperviseDepartRole()) {
            searchStream =
                    searchStream.filter(RdCompanyDocument$.SUPERVISE_DEPART_ID.in(userModel.getAssignedSuperviseDepartIds()));

            final String[] institutionalCodes = BusinessUtil.getDefaultInstitutionals(request.getDeskTypes());
            searchStream = searchStream
                    .filter(RdCompanyDocument$.INSTITUTIONAL.in(institutionalCodes));
        } else {
            searchStream =
                    searchStream.filter(RdCompanyDocument$.ID.eq(userModel.getOrganizationId()));
        }

        if (StringUtils.hasLength(request.getName())) {
            searchStream = searchStream.filter(RdCompanyDocument$.NAME.startsWith(request.getName()));
        }

        if (request.getHasDesk() != null) {
            if (request.getHasDesk()) {
                searchStream = searchStream.filter(RdCompanyDocument$.COUNT_OF_DESK.gt(0));
            } else {
                searchStream = searchStream.filter(RdCompanyDocument$.COUNT_OF_DESK.le(0));
            }
        }

        if (request.getHasPoints() != null) {
            if (request.getHasPoints()) {
                searchStream = searchStream.filter(RdCompanyDocument$.POINTS.gt(0l));
            } else {
                searchStream = searchStream.filter(RdCompanyDocument$.POINTS.le(0l));
            }
        }

        if (request.getHasLocation() != null) {
            if (request.getHasLocation()) {
                searchStream = searchStream.filter(RdCompanyDocument$.HAS_LON_LAT_FLAG.eq(1));
            } else {
                searchStream = searchStream.filter(RdCompanyDocument$.HAS_LON_LAT_FLAG.eq(0));
            }
        }

        int pageIndex = request.getPageIndex();
        if (pageIndex < 1) {
            pageIndex = 1;
        }

        int skip = (pageIndex - 1) * request.getPageSize();

        long totalCount = searchStream.count();
        Collection<RdCompanyDocument> companyDocuments =
                searchStream.skip(skip).limit(request.getPageSize()).collect(Collectors.toList());

        Collection<CompanyAdvanceResponse> data
                = companyDocuments.stream().map(ii -> {
            Double lat = null;
            Double lon = null;
            if (ii.getLonLat() != null) {
                lat = ii.getLonLat().getY();
                lon = ii.getLonLat().getX();
            }

            CompanyAdvanceResponse rs =
                    CompanyAdvanceResponse.create(ii.getId(), ii.getName(),
                            ii.getLegal(), ii.getLegalTelephone(),
                            ii.getPoints(), ii.getCountOfDesk(),ii.getCountOfTemporaryProject(), lat, lon, ii.getHasLonLatFlag() == 1, ii.getHasLonLatFlag() == 1);

            return rs;
        }).collect(Collectors.toList());


        data = buildWithStation(data, userModel);

        return EntityCollection.create(data, request.getPageSize(), totalCount);
    }

    @Override
    public Collection<SearchNearByCompanyResponse> search(SearchNearByCompanyRequest request) {
        /**
         * 先禁用此功能
         */
        if (CollectionUtils.isEmpty(request.getLocations()) || true) {
            return Collections.EMPTY_LIST;
        }

        Collection<SearchNearByCompanyResponse> data = new ArrayList<>();
        for (LonLatValueType lt : request.getLocations()) {
            SearchNearByCompanyResponse companyResponse =
                    getNearestCompany(request.getInstitutional(), lt, request.getDistance());

            if (companyResponse != null) {
                data.add(companyResponse);
            }
        }

        return data;
    }
    
    private Collection<CompanyAdvanceResponse> buildWithStation(
            Collection<CompanyAdvanceResponse> data,
            TenantEmployeeContext.TenantUserModel userModel
    ) {
        if (CollectionUtils.isEmpty(data)) {
            return data;
        }

        Collection<String> companyIds = data.stream().map(ii -> ii.getId()).collect(Collectors.toList());

        SearchStream<RdSecurityStationDocument> stationDocumentStream =
                this.entityStream.of(RdSecurityStationDocument.class);
        if (userModel.isSuperviseDepartRole()) {

            String[] superviseDepartIds = userModel.getAssignedSuperviseDepartIds();
            stationDocumentStream =
                    stationDocumentStream.filter(RdSecurityStationDocument$.SCOPE_PERMISSIONS.in(superviseDepartIds));

            stationDocumentStream =
                    stationDocumentStream
                            .filter(RdSecurityStationDocument$.ORGANIZATION_ID.in(companyIds.toArray(ii -> new String[ii])));
        } else {

            if (!companyIds.contains(userModel.getOrganizationId())) {
                return data;
            }
            stationDocumentStream =
                    stationDocumentStream.filter(RdSecurityStationDocument$.ORGANIZATION_ID.eq(userModel.getOrganizationId()));
        }

        AggregationStream aggregationStream =
                stationDocumentStream
                        .groupBy(RdSecurityStationDocument$.ORGANIZATION_ID)
                        /**
                         * 如果CountOFSecurityMan没有设置为@Index, 那么该表达式无效
                         */
                        .reduce(ReducerFunction.SUM,RdSecurityStationDocument$.COUNT_OF_SECURITY_MAN).as("countOfSecurityMan")
                        .reduce(ReducerFunction.COUNT).as("cntOfStation")
  ;

        List<Triple<String, Integer, Long>> stationResult = aggregationStream.toList(String.class, Integer.class, Long.class);

        for (CompanyAdvanceResponse item : data) {
            Triple<String, Integer, Long> selectedTriple =
                    stationResult.stream().filter(ii -> ii.getFirst().equalsIgnoreCase(item.getId()))
                            .findFirst().orElse(null);
            if (selectedTriple != null) {
                item.assignStationInfo(selectedTriple.getThird(), selectedTriple.getSecond());
            }
        }

        return data;
    }

    
    private static SearchStream<RdCompanyDocument> filterByKeyword(
            SearchStream<RdCompanyDocument> searchStream,
            String input) {
        if (StringUtils.hasLength(input)) {
            String searchKeyword = input.trim();
            String expression = KeywordTermUtils.getFilterExpression(searchKeyword);
            searchStream = searchStream.filter(expression);
        }

        return searchStream;
    }

        /**
     * FT.AGGREGATE com.bcxin.tenant.open.document.domains.documents.RdCompanyDocumentIdx "*" LOAD 1 @lonLat FILTER "exists(@lonLat)" APPLY "geodistance(@lonLat,118.051111,24.606272)" AS dist SORTBY 2 @dist asc
     * <p>
     * 查询出符合条件的坐标和Id
     * FT.AGGREGATE com.bcxin.tenant.open.document.domains.documents.RdCompanyDocumentIdx "*" LOAD 2 @lonLat @id APPLY "geodistance(@lonLat,118.051111,24.606272)" AS dist SORTBY 2 @dist asc
     *
     * @param institutional
     * @param location
     * @param distance
     * @return
     */
    private SearchNearByCompanyResponse getNearestCompany(
            Collection<String> institutional, LonLatValueType location,
            double distance) {
        SearchStream<RdCompanyDocument> companyDocumentSearchStream =
                this.entityStream.of(RdCompanyDocument.class);


        if (!CollectionUtils.isEmpty(institutional)) {
            companyDocumentSearchStream =
                    companyDocumentSearchStream
                            .filter(RdCompanyDocument$.INSTITUTIONAL.in(institutional.toArray(i -> new String[i])));
        }
        companyDocumentSearchStream =
                companyDocumentSearchStream.filter(
                        RdCompanyDocument$.LON_LAT.near(new Point(location.getLon(), location.getLat()),
                                new Distance(distance, Metrics.METERS)
                        ));

        AggregationStream aggregationStream =
                companyDocumentSearchStream
                        .load(RdCompanyDocument$.ID, RdCompanyDocument$.LON_LAT, RdCompanyDocument$.NAME)
                        .apply(String.format("geodistance(@lonLat,%s,%s)", location.getLon(), location.getLat()), "dist")
                        .sorted(Sort.Order.asc("@dist"))
                        .limit(1);

        List<QuadImpl<String, String, String, Double>> companyNearDocuments =
                aggregationStream.toList(String.class, String.class, String.class, Double.class);
        if (CollectionUtils.isEmpty(companyNearDocuments)) {
            return null;
        }

        QuadImpl<String, String, String, Double> selectedQuad = companyNearDocuments.stream().findFirst().get();

        return SearchNearByCompanyResponse.create(selectedQuad.getFirst(), selectedQuad.getSecond(), location.getLat(),
                location.getLon(), selectedQuad.getFourth()
        );
    }
    
}
